@@ -85,7 +85,10 @@
 #'     we use zero as the mode (\code{mode = "zero"}) or estimate it using
 #'     the procedure of Robertson and Cryer (1974)
 #'     (\code{mode = "estimate"})?
-#'
+#' @param win A positive integer. The window size. This will constrain the
+#'     correlations calculated to those +/- the window size. This will
+#'     only improve speed if the window size is \emph{much} less than the
+#'     number of SNPs.
 #'
 #' @seealso
 #' \describe{
@@ -102,6 +105,11 @@
 #'   \item{\code{rr}}{The estimated reliability ratio for each SNP. This
 #'       is the multiplicative factor applied to the naive LD estimate
 #'       for each SNP.}
+#'   \item{\code{rr_raw}}{The raw reliability ratios (for the covariance,
+#'       not the correlation). Only returned if \code{shrinkrr = TRUE}.}
+#'   \item{\code{rr_se}}{The standard errors for the \emph{log}-raw
+#'       reliability ratios for each SNP. That is, we have
+#'       sd(log(rr_raw)) ~ rr_se. Only returned if \code{shrinkrr = TRUE}.}
 #'   \item{\code{semat}}{A matrix of standard errors of the corresponding
 #'       estimators of LD.}
 #' }
@@ -140,7 +148,8 @@ ldfast <- function(gp,
                    se = TRUE,
                    thresh = TRUE,
                    upper = 10,
-                   mode = c("zero", "estimate")) {
+                   mode = c("zero", "estimate"),
+                   win = NULL) {
   ## Check input -------------------------------------------------------------
   stopifnot(inherits(gp, "array"))
   stopifnot(length(dim(gp)) == 3)
@@ -152,6 +161,9 @@ ldfast <- function(gp,
   stopifnot(length(thresh) == 1)
   stopifnot(is.numeric(upper))
   stopifnot(length(numeric) == 1)
+  if (!is.null(win)) {
+    stopifnot(length(win) == 1, win > 0, win <= dim(gp)[[1]])
+  }
   type <- match.arg(type)
   mode <- match.arg(mode)
 
@@ -170,7 +182,8 @@ ldfast <- function(gp,
 
   ## Calculate reliability ratios ---------------------------------------------
   ## after this step rr should be for correlation, *not* for covariance
-  rr <- (muy + varx) / varx
+  rr_raw <- (muy + varx) / varx
+  rr_raw[is.nan(rr_raw)] <- NA_real_
   if (shrinkrr) {
     amom <- abind::abind(pm_mat, pm_mat^2, pv_mat, along = 3)
     mbar <- apply(X = amom, MARGIN = 3, FUN = rowMeans, na.rm = TRUE)
@@ -185,7 +198,7 @@ ldfast <- function(gp,
       1 / (mbar[, 2] - mbar[, 1]^2)
     gradmat[, 3] <-
       1 / (mbar[, 3] + mbar[, 2] - mbar[, 1]^2)
-    svec <- rep(NA_real_, nsnp) # Variances for log rr
+    svec <- rep(NA_real_, nsnp) # Variances for log rr_raw
     nvec <- rowSums(!is.na(pm_mat)) # missing values
     for (i in seq_len(nsnp)) {
       svec[[i]] <- gradmat[i, , drop = FALSE] %*% covarray[, , i, drop = TRUE] %*% t(gradmat[i, , drop = FALSE])
@@ -193,9 +206,9 @@ ldfast <- function(gp,
     svec <- svec / nvec
     svec[svec < 0] <- 0
     if (thresh) {
-      svec[rr > upper] <- Inf
+      svec[rr_raw > upper] <- Inf
     }
-    lvec <- log(rr)
+    lvec <- log(rr_raw)
     if (mode == "estimate") {
       modest <- modeest::hsm(x = lvec)
     } else if (mode == "zero") {
@@ -207,6 +220,7 @@ ldfast <- function(gp,
                         mode = modest)
     rr <- exp(ashr::get_pm(a = ashout) / 2) ## divide by 2 for square root
   } else {
+    rr <- rr_raw
     if (thresh) {
       rr[rr > upper] <- stats::median(rr)
     }
@@ -214,7 +228,12 @@ ldfast <- function(gp,
   }
 
   ## Calculate Pearson correlation --------------------------------------------
-  ldmat <- rr * stats::cor(t(pm_mat), use = "pairwise.complete.obs") * rep(rr, each = nsnp)
+  if (is.null(win)) {
+    ldmat <- rr * stats::cor(t(pm_mat), use = "pairwise.complete.obs") * rep(rr, each = nsnp)
+  } else {
+    ldmat <- rr * slcor(t(pm_mat), win = win) * rep(rr, each = nsnp)
+  }
+
   if (type != "Dprime") {
     if (se) {
       which_truncate <- (ldmat > 1) | (ldmat < -1)
@@ -239,9 +258,8 @@ ldfast <- function(gp,
     deltam_pos <- pmin(outer(X = mux, Y = ploidy - mux, FUN = `*`),
                        outer(X = ploidy - mux, Y = mux, FUN = `*`)) / ploidy ^ 2
 
-    ldmat[ldmat < 0] <- ldmat[ldmat < 0] / deltam_neg[ldmat < 0]
-    ldmat[ldmat > 0] <- ldmat[ldmat > 0] / deltam_pos[ldmat > 0]
-
+    ldmat[ldmat < 0 & !is.na(ldmat)] <- ldmat[ldmat < 0 & !is.na(ldmat)] / deltam_neg[ldmat < 0 & !is.na(ldmat)]
+    ldmat[ldmat > 0 & !is.na(ldmat)] <- ldmat[ldmat > 0 & !is.na(ldmat)] / deltam_pos[ldmat > 0 & !is.na(ldmat)]
 
     if (se) {
       which_truncate <- (ldmat > ploidy) | (ldmat < -ploidy)
@@ -261,6 +279,12 @@ ldfast <- function(gp,
   ## Return list --------------------------------------------------------------
   retlist <- list(ldmat = ldmat, rr = rr)
 
+  ## Add standard errors of rr if shrinkrr = TRUE
+  if (shrinkrr) {
+    retlist$rr_raw <- rr_raw
+    retlist$rr_se <- sqrt(svec)
+  }
+
   ## Standard error calculations if option ------------------------------------
   if (se) {
     if (type == "D") {
